// Example "text" module described in the Module Developer's Guide.
syntax = "proto2";
import "yara.proto";

package python;

// This section is required for all protos that define a YARA module.
option (yara.module_options) = {
  // This is the module's name, the one used when importing the module in a
  // YARA rule.
  name : "rhai"
  // The module's root structure is defined by the `Text` module, which can
  // be found below.
  root_message: "python.Python"
  // The Rust module implementing this YARA module is named `text`. It can
  // be found in `src/modules/text.rs`.
  rust_module: "python"
  // The feature that controls whether this module is compiled or not is named
  // `text-module`.
  cargo_feature: "python-module"
};

// This is the module's root structure.
message Python {
  optional bytes scanned = 1;
}


